
#include <stlsoft_new_allocator.h>
#include <stlsoft_malloc_allocator.h>
#include <stlsoft/contract_violation.hpp>


#if 0
void f()
{
	typedef int (*PfnMathsOp)(int arg0, int arg1);

	void	*pfn;

	int      res = (*(PfnMathsOp)pfn)(10, 20);
}

#include <time.h>

void f2()
{
    ::srand((unsigned int)::time(NULL));
    if(::rand() > (RAND_MAX / 2))
    {
    }
}
#endif /* 0 */

// If the tracing code is removed, the class should be exception safe
// and, since releasing is supposed to not throw, the class may very
// well be used as an RAII class
//
// Calling new or malloc depends on the allocator parameter used
// Each instance is built by inheriting a smaller size than requested
//
// I assumed it'd be a good idea to have Size be a multiple of Scale
// hence the static_assert, if it's not:
// - the static_assert can be removed
// - the partial specialization should be for 0 instead of 1 and rely
//   on integer division eventually yielding 0
//
// I thought this would be a better solution (even if it's not as portable)
// as having to write a lot of malloc-s and new-s along with their matching
// releases
//
// The class inherits privately, as it shouldn't be convertible to anything
template<
	  typename Allocator
	, unsigned Size
	, unsigned Scale
>
class parachute
	: private parachute<
		  Allocator
		, Size / Scale
		, Scale
	>
{
public:
	typedef Allocator							allocator_type;
	typedef typename Allocator::pointer			pointer_type;
	typedef pointer_type						holder_type;

	enum
	{
		  size		= Size
		, scale		= Scale
		, next_size	= size / scale
	};

	typedef parachute<
		  allocator_type
		, size
		, scale
	>											class_type;

	typedef parachute<
		  allocator_type
		, next_size
		, scale
	>											base_type;

	parachute(allocator_type alloc = allocator_type())
		: base_type(alloc)
		, m_alloc(alloc)
		, m_holder(m_alloc.allocate(size))
		, m_release(true)
	{
		stlsoft_static_assert((size % Scale) == 0);

		std::cout
			<< "Allocated "
			<< size
			<< " bytes"
			<< "\n"
			;
	}

	~parachute()
	{
		std::cout << "Destructor: ";

		if( m_release )
			free();
		else
		{
			std::cout
				<< size
				<< " bytes were already released\n"
				;
		}
	}

	// Recursively free all of our base class' memory
	void release()
	{
		stlsoft_assert(m_release);

		free();
		base_type::release();
	}

private:
	// Free our own memory only
	void free()
	{
		m_alloc.deallocate(m_holder);
		m_release = false;

		std::cout
			<< "Released "
			<< size
			<< " bytes"
			<< "\n"
			;
	}

	allocator_type	m_alloc;
	holder_type		m_holder;
	bool			m_release;

private:
	// Not to be implemented
	parachute(parachute const&);
	parachute& operator=(parachute const&);
};

// Do nothing implementation
// terminates the inheritance hierarchy
template<
	  typename Allocator
// Size is partially specialized
//	, unsigned Size
	, unsigned Scale
>
class parachute<Allocator,1,Scale>
{
public:
	parachute()
	{
	}

	template<typename T>
	parachute(T)
	{
	}

	void release()
	{
	}
};


#include <stdexcept>
#include <iostream>

int main_(int argc, char **argv, bool should_throw)
{
	typedef ::stlsoft::new_allocator<char>		allocator_t0;
	parachute<
		  allocator_t0
		, 100000
		, 10
	> par1;

	typedef ::stlsoft::malloc_allocator<char>	allocator_t1;
	parachute<
		  allocator_t1
		, 100000
		, 10
	> par2;

	try
	{
		// . . . all the business of the application . . . 
		if( should_throw )
		{
			throw ::std::bad_alloc();
		}
	}
	catch(::std::bad_alloc const& x)
	{
		// Deploy the parachutes with fingers crossed that it leaves enough memory 
		// . . . all the actions you want to carry out to (attempt to) effect orderly shutdown . . .
		std::cout
			<< "Caught an std::bad_alloc exception\n"
			<< x.what()
			<< "\n"
			;

		par1.release();
		par2.release();

		// Test double releasing
		//par1.release();
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}

int main(int argc, char **argv)
{
	// The 'generic' main - carries out a minimum of things, minimising risk
	// of failing to be in a position to 

	try
	{
		main_(argc, argv, false);
		main_(argc, argv, true);
	}
	catch(::std::exception &x)
	{
		fprintf(stderr, "Unhandled error: %s\n", x.what());
	}
	catch(...)
	{
		fprintf(stderr, "Unhandled unknown error\n");
	}

	std::cin.get();
	return EXIT_FAILURE;
}